;
; Extended memory driver for the PLUS256K memory expansion
; Marco van den Heuvel, 2010-01-26
;

	.include	"zeropage.inc"

	.include	"em-kernel.inc"
	.include	"em-error.inc"


	.macpack	generic


; ------------------------------------------------------------------------
; Header. Includes jump table

.segment	"JUMPTABLE"

; Driver signature

	.byte	$65, $6d, $64		; "emd"
	.byte	EMD_API_VERSION		; EM API version number

; Jump table.

	.word	INSTALL
	.word	UNINSTALL
	.word	PAGECOUNT
	.word	MAP
	.word	USE
	.word	COMMIT
	.word	COPYFROM
	.word	COPYTO

; ------------------------------------------------------------------------
; Constants

BASE			= $1000
PAGES			= ($10000 - BASE) / 256 * 3
CHECKPLUS256K	= $0200
FETCHPLUS256K	= $0200
FETCHPLUS256KVEC	= $020B
STASHPLUS256K	= $0212
STASHPLUS256KVEC	= $021F

; ------------------------------------------------------------------------
; Data.

.data

curpage:	.res	2		; Current page number
curbank:	.res	1		; Current bank
backup:	.res	38		; Backup area of data in the location
					; where the copy routine will be
plus256kcheckcode:
	.byte $a2, $40		; ldx #$40
	.byte $a0, $00		; ldy #$00
	.byte	$8c, $00, $d1	; sty $d100
	.byte $ad, $00, $40	; lda $4000
	.byte $8e, $00, $d1	; stx $d100
	.byte $cd, $00, $40	; cmp $4000
	.byte $d0, $0f		; bne present
	.byte $ee, $00, $40	; inc $4000
	.byte $8c, $00, $d1	; sty $d100
	.byte $cd, $00, $40	; cmp $4000
	.byte $f0, $04		; beq present
	.byte $ce, $00, $40	; dec $4000
	.byte	$60			; rts

; present
	.byte	$c8			; iny
	.byte	$60			; rts


plus256kcopycode:
	.byte $8e, $00, $d1	; stx $d100
	.byte $a5, $01		; lda $01
	.byte $aa			; tax
	.byte $29, $f8		; and #$f8
	.byte $85, $01		; sta $01
	.byte $b1, $00		; lda ($xx),y
	.byte $86, $01		; stx $01
	.byte $8c, $00, $d1	; sty $d100
	.byte $60			; rts

	.byte $8e, $00, $d1	; stx $d100
	.byte $48			; pha
	.byte $a5, $01		; lda $01
	.byte $aa			; tax
	.byte $29, $f8		; and #$f8
	.byte $85, $01		; sta $01
	.byte $68			; pla
	.byte $91, $00		; sta ($xx),y
	.byte $86, $01		; stx $01
	.byte $8c, $00, $d1	; sty $d100
	.byte $60			; rts
 
window:	.res	256		; Memory "window"

.code

; ------------------------------------------------------------------------
; INSTALL routine. Is called after the driver is loaded into memory. If
; possible, check if the hardware is present and determine the amount of
; memory available.
; Must return an EM_ERR_xx code in a/x.
;

INSTALL:
	jsr	backup_and_setup_check_routine
	jsr	CHECKPLUS256K
	cli
	ldx	#$22
	jsr	restore_data
	cpy	#$01
	beq	@present
	lda	#<EM_ERR_NO_DEVICE
	ldx	#>EM_ERR_NO_DEVICE
	rts

@present:
	lda	#<EM_ERR_OK
	ldx	#>EM_ERR_OK
;	rts				; Run into UNINSTALL instead

; ------------------------------------------------------------------------
; UNINSTALL routine. Is called before the driver is removed from memory.
; Can do cleanup or whatever. Must not return anything.
;

UNINSTALL:
	rts


; ------------------------------------------------------------------------
; PAGECOUNT: Return the total number of available pages in a/x.
;

PAGECOUNT:
	lda	#<PAGES
	ldx	#>PAGES
	rts

; ------------------------------------------------------------------------
; MAP: Map the page in a/x into memory and return a pointer to the page in
; a/x. The contents of the currently mapped page (if any) may be discarded
; by the driver.
;

MAP:
	sei
	sta	curpage		; Remember the new page
	stx	curpage+1
	jsr	adjust_page_and_bank
	stx	curbank
	clc
	adc	#>BASE
	sta	ptr1+1
	ldy	#0
	sty	ptr1
	jsr	backup_and_setup_copy_routine
	ldx	#<ptr1
      stx	FETCHPLUS256KVEC
@L1:
	ldx	curbank
	jsr	FETCHPLUS256K
	ldx	ptr1
	sta	window,x
	inc	ptr1
	bne	@L1

; Return the memory window

	ldx	#$25
	jsr	restore_data
	lda	#<window
	ldx	#>window		; Return the window address
	cli
	rts

; ------------------------------------------------------------------------
; USE: Tell the driver that the window is now associated with a given page.

USE:
	sta	curpage		; Remember the page
	stx	curpage+1
	lda	#<window
	ldx	#>window		; Return the window
	rts

; ------------------------------------------------------------------------
; COMMIT: Commit changes in the memory window to extended storage.

COMMIT:
	sei
	lda	curpage		; Get the current page
	ldx	curpage+1

	jsr	adjust_page_and_bank
	stx	curbank
	clc
	adc	#>BASE
	sta	ptr1+1
	ldy	#0
	sty	ptr1
	jsr	backup_and_setup_copy_routine
	ldx	#<ptr1
      stx	STASHPLUS256KVEC
@L1:
	ldx	ptr1
	lda	window,x
	ldx	curbank
	jsr	STASHPLUS256K
	inc	ptr1
	bne	@L1

; Return the memory window

	ldx	#$25
	jsr	restore_data
done:
	cli
	rts

; ------------------------------------------------------------------------
; COPYFROM: Copy from extended into linear memory. A pointer to a structure
; describing the request is passed in a/x.
; The function must not return anything.
;


COPYFROM:
	sei
	jsr	setup
	jsr	backup_and_setup_copy_routine

; Setup is:
;
;   - ptr1 contains the struct pointer
;   - ptr2 contains the linear memory buffer
;   - ptr3 contains -(count-1)
;   - ptr4 contains the page memory buffer plus offset
;   - tmp1 contains zero (to be used for linear memory buffer offset)
;   - tmp2 contains the bank value

	lda	#<ptr4
      sta	FETCHPLUS256KVEC
	jmp	@L3

@L1:
	ldx	tmp2
	ldy	#0
	jsr	FETCHPLUS256K
	ldy	tmp1
	sta	(ptr2),y
	inc	tmp1
	bne	@L2
	inc	ptr2+1
@L2:
	inc	ptr4
	beq	@L4

; Bump count and repeat

@L3:
	inc	ptr3
	bne	@L1
	inc	ptr3+1
	bne	@L1
	ldx	#$25
	jsr	restore_data
	cli
	rts

; Bump page register

@L4:
	inc	ptr4+1
	bne	@L3
	lda	#>BASE
	sta	ptr4+1
	lda	tmp2
	clc
	adc	#$40
	sta	tmp2
	jmp	@L3

; ------------------------------------------------------------------------
; COPYTO: Copy from linear into extended memory. A pointer to a structure
; describing the request is passed in a/x.
; The function must not return anything.
;

COPYTO:
	sei
	jsr	setup
	jsr	backup_and_setup_copy_routine

; Setup is:
;
;   - ptr1 contains the struct pointer
;   - ptr2 contains the linear memory buffer
;   - ptr3 contains -(count-1)
;   - ptr4 contains the page memory buffer plus offset
;   - tmp1 contains zero (to be used for linear memory buffer offset)
;   - tmp2 contains the bank value

	lda	#<ptr4
      sta	STASHPLUS256KVEC
	jmp	@L3

@L1:
	ldy	tmp1
	lda	(ptr2),y
	ldx	tmp2
	ldy	#0
	jsr	STASHPLUS256K
	inc	tmp1
	bne	@L2
	inc	ptr2+1
@L2:
	inc	ptr4
	beq	@L4

; Bump count and repeat

@L3:
	inc	ptr3
	bne	@L1
	inc	ptr3+1
	bne	@L1
	ldx	#$25
	jsr	restore_data
	cli
	rts

; Bump page register

@L4:
	inc	ptr4+1
	bne	@L3
	lda	#>BASE
	sta	ptr4+1
	lda	tmp2
	clc
	adc	#$40
	sta	tmp2
	jmp	@L3

; ------------------------------------------------------------------------
; Helper function for COPYFROM and COPYTO: Store the pointer to the request
; structure and prepare data for the copy

setup:
	sta	ptr1
	stx	ptr1+1		; Save passed pointer

; Get the page number from the struct and adjust it so that it may be used
; with the hardware. That is: ptr4 has the page address and page offset
; tmp2 will hold the bank value

	ldy	#EM_COPY::PAGE+1
	lda	(ptr1),y
	tax
	ldy	#EM_COPY::PAGE
	lda	(ptr1),y
	jsr	adjust_page_and_bank
	clc
	adc	#>BASE
	sta	ptr4+1
	stx	tmp2

; Get the buffer pointer into ptr2

	ldy	#EM_COPY::BUF
	lda	(ptr1),y
	sta	ptr2
	iny
	lda	(ptr1),y
	sta	ptr2+1

; Get the count, calculate -(count-1) and store it into ptr3

	ldy	#EM_COPY::COUNT
	lda	(ptr1),y
	eor	#$FF
	sta	ptr3
	iny
	lda	(ptr1),y
	eor	#$FF
	sta	ptr3+1

; Get the page offset into ptr4 and clear tmp1

	ldy	#EM_COPY::OFFS
	lda	(ptr1),y
	sta	ptr4
	lda	#0
	sta	tmp1

; Done

	rts

; Helper routines for copying to and from the +256k ram

backup_and_setup_copy_routine:
	ldx	#$25
@L1:
	lda	$0200,x
	sta	backup,x
	lda	plus256kcopycode,x
	sta	$0200,x
	dex
	bpl	@L1
	rts

backup_and_setup_check_routine:
	ldx	#$22
@L1:
	lda	$0200,x
	sta	backup,x
	lda	plus256kcheckcode,x
	sta	$0200,x
	dex
	bpl	@L1
	rts

restore_data:
	lda	backup,x
	sta	$0200,x
	dex
	bpl	restore_data
	rts

; Helper routine to correct for the bank and page
adjust_page_and_bank:
	sta	tmp4
	cpx	#2
	beq	@bank3_second_range
	cpx	#1
	beq	@bank2_or_bank3
	sec
	sbc	#$F0
	bcs	@bank2_first_range
	lda	tmp4
	ldx	#$40
	rts

@bank2_or_bank3:
	sec
	sbc	#$E0
	bcs	@bank3_first_range
	lda	tmp4
	clc
	adc	#$10
@bank2_first_range:
	ldx	#$80
	rts

@bank3_second_range:
	clc
	adc	#$20
@bank3_first_range:
	ldx	#$C0
	rts
